// Copyright 2018 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

package com.mapmyindia.mapmyindiagl;

import android.content.Context;
import android.graphics.Point;
import android.util.DisplayMetrics;

import com.mapmyindia.sdk.geoanalytics.GeoAnalyticsAppearanceOption;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsDetail;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsRequest;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsType;
import com.mapmyindia.sdk.geojson.Polygon;
import com.mapmyindia.sdk.maps.MapmyIndiaMap;
import com.mapmyindia.sdk.maps.camera.CameraELocBoundUpdate;
import com.mapmyindia.sdk.maps.camera.CameraELocPosition;
import com.mapmyindia.sdk.maps.camera.CameraELocUpdate;
import com.mapmyindia.sdk.maps.camera.CameraELocUpdateFactory;
import com.mapmyindia.sdk.maps.camera.CameraPosition;
import com.mapmyindia.sdk.maps.camera.CameraUpdate;
import com.mapmyindia.sdk.maps.camera.CameraUpdateFactory;
import com.mapmyindia.sdk.maps.geometry.LatLng;
import com.mapmyindia.sdk.maps.geometry.LatLngBounds;
import com.mapmyindia.sdk.plugin.annotation.Circle;
import com.mapmyindia.sdk.plugin.annotation.Fill;
import com.mapmyindia.sdk.plugin.annotation.Line;
import com.mapmyindia.sdk.plugin.annotation.Symbol;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Conversions between JSON-like values and MapboxMaps data types.
 */
class Convert {

    private final static String TAG = "Convert";

//  private static BitmapDescriptor toBitmapDescriptor(Object o) {
//    final List<?> data = toList(o);
//    switch (toString(data.get(0))) {
//      case "defaultMarker":
//        if (data.size() == 1) {
//          return BitmapDescriptorFactory.defaultMarker();
//        } else {
//          return BitmapDescriptorFactory.defaultMarker(toFloat(data.get(1)));
//        }
//      case "fromAsset":
//        if (data.size() == 2) {
//          return BitmapDescriptorFactory.fromAsset(
//              FlutterMain.getLookupKeyForAsset(toString(data.get(1))));
//        } else {
//          return BitmapDescriptorFactory.fromAsset(
//              FlutterMain.getLookupKeyForAsset(toString(data.get(1)), toString(data.get(2))));
//        }
//      default:
//        throw new IllegalArgumentException("Cannot interpret " + o + " as BitmapDescriptor");
//    }
//  }

    static boolean toBoolean(Object o) {
        return (Boolean) o;
    }

    static CameraPosition toCameraPosition(Object o) {
        final Map<?, ?> data = toMap(o);
        final CameraPosition.Builder builder = new CameraPosition.Builder();
        builder.bearing(toFloat(data.get("bearing")));
        builder.target(toLatLng(data.get("target")));
        builder.tilt(toFloat(data.get("tilt")));
        builder.zoom(toFloat(data.get("zoom")));
        return builder.build();
    }

    static List<String> toAnnotationOrder(Object o) {
        final List<?> data = toList(o);
        List<String> annotations = new ArrayList();
        for (int index = 0; index < data.size(); index++) {
            annotations.add(toString(data.get(index)));
        }
        return annotations;
    }

    static List<String> toAnnotationConsumeTapEvents(Object o) {
        return toAnnotationOrder(o);
    }

    static boolean isScrollByCameraUpdate(Object o) {
        return toString(toList(o).get(0)).equals("scrollBy");
    }

    static CameraUpdate toCameraUpdate(Object o, MapmyIndiaMap mapboxMap, float density) {
        final List<?> data = toList(o);
        switch (toString(data.get(0))) {
            case "newCameraPosition":
                return CameraUpdateFactory.newCameraPosition(toCameraPosition(data.get(1)));
            case "newLatLng":
                return CameraUpdateFactory.newLatLng(toLatLng(data.get(1)));
            case "newLatLngBounds":
                return CameraUpdateFactory.newLatLngBounds(toLatLngBounds(data.get(1)), toPixels(data.get(2), density),
                        toPixels(data.get(3), density), toPixels(data.get(4), density), toPixels(data.get(5), density));
            case "newLatLngZoom":
                return CameraUpdateFactory.newLatLngZoom(toLatLng(data.get(1)), toFloat(data.get(2)));
            case "scrollBy":
                mapboxMap.scrollBy(
                        toFractionalPixels(data.get(1), density),
                        toFractionalPixels(data.get(2), density)
                );
                return null;
            case "zoomBy":
                if (data.size() == 2) {
                    return CameraUpdateFactory.zoomBy(toFloat(data.get(1)));
                } else {
                    return CameraUpdateFactory.zoomBy(toFloat(data.get(1)), toPoint(data.get(2), density));
                }
            case "zoomIn":
                return CameraUpdateFactory.zoomIn();
            case "zoomOut":
                return CameraUpdateFactory.zoomOut();
            case "zoomTo":
                return CameraUpdateFactory.zoomTo(toFloat(data.get(1)));
            case "bearingTo":
                return CameraUpdateFactory.bearingTo(toFloat(data.get(1)));
            case "tiltTo":
                return CameraUpdateFactory.tiltTo(toFloat(data.get(1)));
            default:
                throw new IllegalArgumentException("Cannot interpret " + o + " as CameraUpdate");
        }
    }

    static double toDouble(Object o) {
        return ((Number) o).doubleValue();
    }

    static float toFloat(Object o) {
        return ((Number) o).floatValue();
    }

    static Float toFloatWrapper(Object o) {
        return (o == null) ? null : toFloat(o);
    }

    static int toInt(Object o) {
        return ((Number) o).intValue();
    }

    static Object toJson(CameraPosition position) {
        if (position == null) {
            return null;
        }
        final Map<String, Object> data = new HashMap<>();
        data.put("bearing", position.bearing);
        data.put("target", toJson(position.target));
        data.put("tilt", position.tilt);
        data.put("zoom", position.zoom);
        return data;
    }

    private static Object toJson(LatLng latLng) {
        return Arrays.asList(latLng.getLatitude(), latLng.getLongitude());
    }

    static LatLng toLatLng(Object o) {
        final List<?> data = toList(o);
        return new LatLng(toDouble(data.get(0)), toDouble(data.get(1)));
    }

    static LatLngBounds toLatLngBounds(Object o) {
        if (o == null) {
            return null;
        }
        final List<?> data = toList(o);
        LatLng[] boundsArray = new LatLng[]{toLatLng(data.get(0)), toLatLng(data.get(1))};
        List<LatLng> bounds = Arrays.asList(boundsArray);
        LatLngBounds.Builder builder = new LatLngBounds.Builder();
        builder.includes(bounds);
        return builder.build();
    }

    static List<LatLng> toLatLngList(Object o, boolean flippedOrder) {
        if (o == null) {
            return null;
        }
        final List<?> data = toList(o);
        List<LatLng> latLngList = new ArrayList<>();
        for (int i = 0; i < data.size(); i++) {
            final List<?> coords = toList(data.get(i));
            if (flippedOrder) {
                latLngList.add(new LatLng(toDouble(coords.get(1)), toDouble(coords.get(0))));
            } else {
                latLngList.add(new LatLng(toDouble(coords.get(0)), toDouble(coords.get(1))));
            }
        }
        return latLngList;
    }

    private static List<List<LatLng>> toLatLngListList(Object o) {
        if (o == null) {
            return null;
        }
        final List<?> data = toList(o);
        List<List<LatLng>> latLngListList = new ArrayList<>();
        for (int i = 0; i < data.size(); i++) {
            List<LatLng> latLngList = toLatLngList(data.get(i), false);
            latLngListList.add(latLngList);
        }
        return latLngListList;
    }

    static Polygon interpretListLatLng(List<List<LatLng>> geometry) {
        List<List<com.mapmyindia.sdk.geojson.Point>> points = new ArrayList<>(geometry.size());
        for (List<LatLng> innerGeometry : geometry) {
            List<com.mapmyindia.sdk.geojson.Point> innerPoints = new ArrayList<>(innerGeometry.size());
            for (LatLng latLng : innerGeometry) {
                innerPoints.add(com.mapmyindia.sdk.geojson.Point.fromLngLat(latLng.getLongitude(), latLng.getLatitude()));
            }
            points.add(innerPoints);
        }
        return Polygon.fromLngLats(points);
    }

    static List<?> toList(Object o) {
        return (List<?>) o;
    }

    static long toLong(Object o) {
        return ((Number) o).longValue();
    }

    static Map<?, ?> toMap(Object o) {
        return (Map<?, ?>) o;
    }

    private static float toFractionalPixels(Object o, float density) {
        return toFloat(o) * density;
    }

    static int toPixels(Object o, float density) {
        return (int) toFractionalPixels(o, density);
    }

    private static Point toPoint(Object o, float density) {
        final List<?> data = toList(o);
        return new Point(toPixels(data.get(0), density), toPixels(data.get(1), density));
    }

    static String toString(Object o) {
        return (String) o;
    }

    static void interpretMapboxMapOptions(Object o, MapboxMapOptionsSink sink, Context context) {
        final DisplayMetrics metrics = context.getResources().getDisplayMetrics();
        final Map<?, ?> data = toMap(o);
        final Object cameraTargetBounds = data.get("cameraTargetBounds");
        if (cameraTargetBounds != null) {
            final List<?> targetData = toList(cameraTargetBounds);
            sink.setCameraTargetBounds(toLatLngBounds(targetData.get(0)));
        }
        final Object compassEnabled = data.get("compassEnabled");
        if (compassEnabled != null) {
            sink.setCompassEnabled(toBoolean(compassEnabled));
        }
        final Object styleString = data.get("styleString");
        if (styleString != null) {
            sink.setStyleString(toString(styleString));
        }
        final Object mapmyIndiaStyle = data.get("mapmyIndiaStyle");
        if (mapmyIndiaStyle != null) {
            sink.setMapmyIndiaStyle(toString(mapmyIndiaStyle));
        }
        final Object minMaxZoomPreference = data.get("minMaxZoomPreference");
        if (minMaxZoomPreference != null) {
            final List<?> zoomPreferenceData = toList(minMaxZoomPreference);
            sink.setMinMaxZoomPreference( //
                    toFloatWrapper(zoomPreferenceData.get(0)), //
                    toFloatWrapper(zoomPreferenceData.get(1)));
        }
        final Object rotateGesturesEnabled = data.get("rotateGesturesEnabled");
        if (rotateGesturesEnabled != null) {
            sink.setRotateGesturesEnabled(toBoolean(rotateGesturesEnabled));
        }
        final Object scrollGesturesEnabled = data.get("scrollGesturesEnabled");
        if (scrollGesturesEnabled != null) {
            sink.setScrollGesturesEnabled(toBoolean(scrollGesturesEnabled));
        }
        final Object tiltGesturesEnabled = data.get("tiltGesturesEnabled");
        if (tiltGesturesEnabled != null) {
            sink.setTiltGesturesEnabled(toBoolean(tiltGesturesEnabled));
        }
        final Object trackCameraPosition = data.get("trackCameraPosition");
        if (trackCameraPosition != null) {
            sink.setTrackCameraPosition(toBoolean(trackCameraPosition));
        }
        final Object zoomGesturesEnabled = data.get("zoomGesturesEnabled");
        if (zoomGesturesEnabled != null) {
            sink.setZoomGesturesEnabled(toBoolean(zoomGesturesEnabled));
        }
        final Object myLocationEnabled = data.get("myLocationEnabled");
        if (myLocationEnabled != null) {
            sink.setMyLocationEnabled(toBoolean(myLocationEnabled));
        }
        final Object myLocationTrackingMode = data.get("myLocationTrackingMode");
        if (myLocationTrackingMode != null) {
            sink.setMyLocationTrackingMode(toInt(myLocationTrackingMode));
        }
        final Object myLocationRenderMode = data.get("myLocationRenderMode");
        if (myLocationRenderMode != null) {
            sink.setMyLocationRenderMode(toInt(myLocationRenderMode));
        }
        final Object logoViewMargins = data.get("logoViewMargins");
        if (logoViewMargins != null) {
            final List logoViewMarginsData = toList(logoViewMargins);
            final Point point = toPoint(logoViewMarginsData, metrics.density);
            sink.setLogoViewMargins(point.x, point.y);
        }
        final Object compassGravity = data.get("compassViewPosition");
        if (compassGravity != null) {
            sink.setCompassGravity(toInt(compassGravity));
        }
        final Object compassViewMargins = data.get("compassViewMargins");
        if (compassViewMargins != null) {
            final List compassViewMarginsData = toList(compassViewMargins);
            final Point point = toPoint(compassViewMarginsData, metrics.density);
            sink.setCompassViewMargins(point.x, point.y);
        }
        final Object attributionButtonGravity = data.get("attributionButtonPosition");
        if (attributionButtonGravity != null) {
            sink.setAttributionButtonGravity(toInt(attributionButtonGravity));
        }
        final Object attributionButtonMargins = data.get("attributionButtonMargins");
        if (attributionButtonMargins != null) {
            final List attributionButtonMarginsData = toList(attributionButtonMargins);
            final Point point = toPoint(attributionButtonMarginsData, metrics.density);
            sink.setAttributionButtonMargins(point.x, point.y);
        }
        final Object showGeoAnalyticsInfoWindow = data.get("showGeoAnalyticsInfoWindow");
        if (showGeoAnalyticsInfoWindow != null) {
            sink.showGeoAnalyticsInfoWindow(toBoolean(showGeoAnalyticsInfoWindow));
        }
    }

    static void interpretSymbolOptions(Object o, SymbolOptionsSink sink) {
        final Map<?, ?> data = toMap(o);
        final Object iconSize = data.get("iconSize");
        if (iconSize != null) {
            sink.setIconSize(toFloat(iconSize));
        }
        final Object iconImage = data.get("iconImage");
        if (iconImage != null) {
            sink.setIconImage(toString(iconImage));
        }
        final Object iconRotate = data.get("iconRotate");
        if (iconRotate != null) {
            sink.setIconRotate(toFloat(iconRotate));
        }
        final Object iconOffset = data.get("iconOffset");
        if (iconOffset != null) {
            sink.setIconOffset(new float[]{toFloat(toList(iconOffset).get(0)), toFloat(toList(iconOffset).get(1))});
        }
        final Object iconAnchor = data.get("iconAnchor");
        if (iconAnchor != null) {
            sink.setIconAnchor(toString(iconAnchor));
        }
        final ArrayList fontNames = (ArrayList) data.get("fontNames");
        if (fontNames != null) {
            sink.setFontNames((String[]) fontNames.toArray(new String[0]));
        }
        final Object textField = data.get("textField");
        if (textField != null) {
            sink.setTextField(toString(textField));
        }
        final Object textSize = data.get("textSize");
        if (textSize != null) {
            sink.setTextSize(toFloat(textSize));
        }
        final Object textMaxWidth = data.get("textMaxWidth");
        if (textMaxWidth != null) {
            sink.setTextMaxWidth(toFloat(textMaxWidth));
        }
        final Object textLetterSpacing = data.get("textLetterSpacing");
        if (textLetterSpacing != null) {
            sink.setTextLetterSpacing(toFloat(textLetterSpacing));
        }
        final Object textJustify = data.get("textJustify");
        if (textJustify != null) {
            sink.setTextJustify(toString(textJustify));
        }
        final Object textAnchor = data.get("textAnchor");
        if (textAnchor != null) {
            sink.setTextAnchor(toString(textAnchor));
        }
        final Object textRotate = data.get("textRotate");
        if (textRotate != null) {
            sink.setTextRotate(toFloat(textRotate));
        }
        final Object textTransform = data.get("textTransform");
        if (textTransform != null) {
            sink.setTextTransform(toString(textTransform));
        }
        final Object textOffset = data.get("textOffset");
        if (textOffset != null) {
            sink.setTextOffset(new float[]{toFloat(toList(textOffset).get(0)), toFloat(toList(textOffset).get(1))});
        }
        final Object iconOpacity = data.get("iconOpacity");
        if (iconOpacity != null) {
            sink.setIconOpacity(toFloat(iconOpacity));
        }
        final Object iconColor = data.get("iconColor");
        if (iconColor != null) {
            sink.setIconColor(toString(iconColor));
        }
        final Object iconHaloColor = data.get("iconHaloColor");
        if (iconHaloColor != null) {
            sink.setIconHaloColor(toString(iconHaloColor));
        }
        final Object iconHaloWidth = data.get("iconHaloWidth");
        if (iconHaloWidth != null) {
            sink.setIconHaloWidth(toFloat(iconHaloWidth));
        }
        final Object iconHaloBlur = data.get("iconHaloBlur");
        if (iconHaloBlur != null) {
            sink.setIconHaloBlur(toFloat(iconHaloBlur));
        }
        final Object textOpacity = data.get("textOpacity");
        if (textOpacity != null) {
            sink.setTextOpacity(toFloat(textOpacity));
        }
        final Object textColor = data.get("textColor");
        if (textColor != null) {
            sink.setTextColor(toString(textColor));
        }
        final Object textHaloColor = data.get("textHaloColor");
        if (textHaloColor != null) {
            sink.setTextHaloColor(toString(textHaloColor));
        }
        final Object textHaloWidth = data.get("textHaloWidth");
        if (textHaloWidth != null) {
            sink.setTextHaloWidth(toFloat(textHaloWidth));
        }
        final Object textHaloBlur = data.get("textHaloBlur");
        if (textHaloBlur != null) {
            sink.setTextHaloBlur(toFloat(textHaloBlur));
        }
        final Object geometry = data.get("geometry");
        if (geometry != null) {
            sink.setGeometry(toLatLng(geometry));
        }
        final Object symbolSortKey = data.get("zIndex");
        if (symbolSortKey != null) {
            sink.setSymbolSortKey(toFloat(symbolSortKey));
        }
        final Object draggable = data.get("draggable");
        if (draggable != null) {
            sink.setDraggable(toBoolean(draggable));
        }
        final Object eLoc = data.get("eLoc");
        if (eLoc != null) {
            sink.setELoc(toString(eLoc));
        }
    }

    static void interpretCircleOptions(Object o, CircleOptionsSink sink) {
        final Map<?, ?> data = toMap(o);
        final Object circleRadius = data.get("circleRadius");
        if (circleRadius != null) {
            sink.setCircleRadius(toFloat(circleRadius));
        }
        final Object circleColor = data.get("circleColor");
        if (circleColor != null) {
            sink.setCircleColor(toString(circleColor));
        }
        final Object circleBlur = data.get("circleBlur");
        if (circleBlur != null) {
            sink.setCircleBlur(toFloat(circleBlur));
        }
        final Object circleOpacity = data.get("circleOpacity");
        if (circleOpacity != null) {
            sink.setCircleOpacity(toFloat(circleOpacity));
        }
        final Object circleStrokeWidth = data.get("circleStrokeWidth");
        if (circleStrokeWidth != null) {
            sink.setCircleStrokeWidth(toFloat(circleStrokeWidth));
        }
        final Object circleStrokeColor = data.get("circleStrokeColor");
        if (circleStrokeColor != null) {
            sink.setCircleStrokeColor(toString(circleStrokeColor));
        }
        final Object circleStrokeOpacity = data.get("circleStrokeOpacity");
        if (circleStrokeOpacity != null) {
            sink.setCircleStrokeOpacity(toFloat(circleStrokeOpacity));
        }
        final Object geometry = data.get("geometry");
        if (geometry != null) {
            sink.setGeometry(toLatLng(geometry));
        }
        final Object draggable = data.get("draggable");
        if (draggable != null) {
            sink.setDraggable(toBoolean(draggable));
        }
    }

    static void interpretLineOptions(Object o, LineOptionsSink sink) {
        final Map<?, ?> data = toMap(o);
        final Object lineJoin = data.get("lineJoin");
        if (lineJoin != null) {
            sink.setLineJoin(toString(lineJoin));
        }
        final Object lineOpacity = data.get("lineOpacity");
        if (lineOpacity != null) {
            sink.setLineOpacity(toFloat(lineOpacity));
        }
        final Object lineColor = data.get("lineColor");
        if (lineColor != null) {
            sink.setLineColor(toString(lineColor));
        }
        final Object lineWidth = data.get("lineWidth");
        if (lineWidth != null) {
            sink.setLineWidth(toFloat(lineWidth));
        }
        final Object lineGapWidth = data.get("lineGapWidth");
        if (lineGapWidth != null) {
            sink.setLineGapWidth(toFloat(lineGapWidth));
        }
        final Object lineOffset = data.get("lineOffset");
        if (lineOffset != null) {
            sink.setLineOffset(toFloat(lineOffset));
        }
        final Object lineBlur = data.get("lineBlur");
        if (lineBlur != null) {
            sink.setLineBlur(toFloat(lineBlur));
        }
        final Object linePattern = data.get("linePattern");
        if (linePattern != null) {
            sink.setLinePattern(toString(linePattern));
        }
        final Object geometry = data.get("geometry");
        if (geometry != null) {
            sink.setGeometry(toLatLngList(geometry, false));
        }
        final Object draggable = data.get("draggable");
        if (draggable != null) {
            sink.setDraggable(toBoolean(draggable));
        }
    }

    static void interpretFillOptions(Object o, FillOptionsSink sink) {
        final Map<?, ?> data = toMap(o);
        final Object fillOpacity = data.get("fillOpacity");
        if (fillOpacity != null) {
            sink.setFillOpacity(toFloat(fillOpacity));
        }
        final Object fillColor = data.get("fillColor");
        if (fillColor != null) {
            sink.setFillColor(toString(fillColor));
        }
        final Object fillOutlineColor = data.get("fillOutlineColor");
        if (fillOutlineColor != null) {
            sink.setFillOutlineColor(toString(fillOutlineColor));
        }
        final Object fillPattern = data.get("fillPattern");
        if (fillPattern != null) {
            sink.setFillPattern(toString(fillPattern));
        }
        final Object geometry = data.get("geometry");
        if (geometry != null) {
            sink.setGeometry(toLatLngListList(geometry));
        }
        final Object draggable = data.get("draggable");
        if (draggable != null) {
            sink.setDraggable(toBoolean(draggable));
        }
    }

    static void moveCameraWithELoc(Object o, MapmyIndiaMap mapmyIndiaMap, MapboxMapController.OnCameraMoveFinishedListener onCameraMoveFinishedListener, float density) {
        final List<?> data = toList(o);
        if ("newELocBounds".equals(toString(data.get(0)))) {
            mapmyIndiaMap.moveCamera(toCameraELocBoundUpdate(o, density), onCameraMoveFinishedListener);
        } else {
            mapmyIndiaMap.moveCamera(toCameraELocUpdate(o), onCameraMoveFinishedListener);
        }
    }

    static void animateCameraWithELoc(Object o, MapmyIndiaMap mapmyIndiaMap, Integer duration, MapmyIndiaMap.CancelableCallback cancelableCallback, float density) {
        final List<?> data = toList(o);
        if ("newELocBounds".equals(toString(data.get(0)))) {
            if (duration != null) {
                mapmyIndiaMap.animateCamera(toCameraELocBoundUpdate(o, density), duration, cancelableCallback);
            } else {
                mapmyIndiaMap.animateCamera(toCameraELocBoundUpdate(o, density), cancelableCallback);
            }
        } else {
            if (duration != null) {
                mapmyIndiaMap.animateCamera(toCameraELocUpdate(o), duration, cancelableCallback);
            } else {
                mapmyIndiaMap.animateCamera(toCameraELocUpdate(o), cancelableCallback);
            }
        }
    }

    static void easeCameraWithELoc(Object o, MapmyIndiaMap mapmyIndiaMap, Integer duration, MapmyIndiaMap.CancelableCallback cancelableCallback, float density) {
        final List<?> data = toList(o);
        if ("newELocBounds".equals(toString(data.get(0)))) {
            if (duration != null) {
                mapmyIndiaMap.easeCamera(toCameraELocBoundUpdate(o, density), duration, cancelableCallback);
            } else {
                mapmyIndiaMap.easeCamera(toCameraELocBoundUpdate(o, density), cancelableCallback);
            }
        } else {
            if (duration != null) {
                mapmyIndiaMap.easeCamera(toCameraELocUpdate(o), duration, cancelableCallback);
            } else {
                mapmyIndiaMap.easeCamera(toCameraELocUpdate(o), cancelableCallback);
            }
        }
    }

    static CameraELocUpdate toCameraELocUpdate(Object o) {
        final List<?> data = toList(o);
        switch (toString(data.get(0))) {
            case "newCameraELocPosition":
                return CameraELocUpdateFactory.newCameraPosition(toCameraELocPosition(data.get(1)));
            case "newELoc":
                return CameraELocUpdateFactory.newELoc(toString(data.get(1)));
            case "newELocZoom":
                return CameraELocUpdateFactory.newELocZoom(toString(data.get(1)), toFloat(data.get(2)));
            default:
                throw new IllegalArgumentException("Cannot interpret " + o + " as CameraUpdate");
        }
    }

    static CameraELocBoundUpdate toCameraELocBoundUpdate(Object o, float density) {
        final List<?> data = toList(o);
        return CameraELocUpdateFactory.newELocBounds(toElocBound(data.get(1)), toPixels(data.get(2), density),
                toPixels(data.get(3), density), toPixels(data.get(4), density), toPixels(data.get(5), density));
    }

    private static List<String> toElocBound(Object o) {
        if (o == null) {
            return null;
        }
        final List<?> data = toList(o);
        List<String> elocBounds = new ArrayList<>();
        for (int i = 0; i < data.size(); i++) {
            elocBounds.add(toString(data.get(i)));
        }
        return elocBounds;
    }

    static CameraELocPosition toCameraELocPosition(Object o) {
        final Map<?, ?> data = toMap(o);
        final CameraELocPosition.Builder builder = new CameraELocPosition.Builder();
        builder.bearing(toFloat(data.get("bearing")));
        builder.target(toString(data.get("target")));
        builder.tilt(toFloat(data.get("tilt")));
        builder.zoom(toFloat(data.get("zoom")));
        return builder.build();
    }

    static Map<String, Object> convertSymbolOptions(String id, Symbol symbol) {
        Map<String, Object> data = new HashMap<>();
        data.put("id", id);
        Map<String, Object> options = new HashMap<>();
        if (symbol.getIconSize() != null) {
            options.put("iconSize", symbol.getIconSize());
        }
        if (symbol.getIconImage() != null) {
            options.put("iconImage", symbol.getIconImage());
        }
        if (symbol.getIconRotate() != null) {
            options.put("iconRotate", symbol.getIconRotate());
        }
        if (symbol.getIconOffset() != null) {
            List<Float> iconOffset = new ArrayList<>();
            iconOffset.add(symbol.getIconOffset().x);
            iconOffset.add(symbol.getIconOffset().y);
            options.put("iconOffset", iconOffset);
        }
        if (symbol.getIconAnchor() != null) {
            options.put("iconAnchor", symbol.getIconAnchor());
        }
        if (symbol.getTextField() != null) {
            options.put("textField", symbol.getTextField());
        }
        if (symbol.getTextSize() != null) {
            options.put("textSize", symbol.getTextSize());
        }
        if (symbol.getTextMaxWidth() != null) {
            options.put("textMaxWidth", symbol.getTextMaxWidth());
        }
        if (symbol.getTextLetterSpacing() != null) {
            options.put("textLetterSpacing", symbol.getTextLetterSpacing());
        }
        if (symbol.getTextJustify() != null) {
            options.put("textJustify", symbol.getTextJustify());
        }
        if (symbol.getTextAnchor() != null) {
            options.put("textAnchor", symbol.getTextAnchor());
        }
        if (symbol.getTextRotate() != null) {
            options.put("textRotate", symbol.getTextRotate());
        }
        if (symbol.getTextTransform() != null) {
            options.put("textTransform", symbol.getTextTransform());
        }
        if (symbol.getTextOffset() != null) {
            List<Float> textOffset = new ArrayList<>();
            textOffset.add(symbol.getTextOffset().x);
            textOffset.add(symbol.getTextOffset().y);
            options.put("textOffset", textOffset);
        }
        if (symbol.getIconOpacity() != null) {
            options.put("iconOpacity", symbol.getIconOpacity());
        }
        if (symbol.getIconColor() != null) {
            options.put("iconColor", symbol.getIconColor());
        }
        if (symbol.getIconHaloColor() != null) {
            options.put("iconHaloColor", symbol.getIconHaloColor());
        }
        if (symbol.getIconHaloWidth() != null) {
            options.put("iconHaloWidth", symbol.getIconHaloWidth());
        }
        if (symbol.getIconHaloBlur() != null) {
            options.put("iconHaloBlur", symbol.getIconHaloBlur());
        }
        if (symbol.getTextOpacity() != null) {
            options.put("textOpacity", symbol.getTextOpacity());
        }
        if (symbol.getTextColor() != null) {
            options.put("textColor", symbol.getTextColor());
        }
        if (symbol.getTextHaloColor() != null) {
            options.put("textHaloColor", symbol.getTextHaloColor());
        }
        if (symbol.getTextHaloWidth() != null) {
            options.put("textHaloWidth", symbol.getTextHaloWidth());
        }
        if (symbol.getTextHaloBlur() != null) {
            options.put("textHaloBlur", symbol.getTextHaloBlur());
        }
        if (symbol.getSymbolSortKey() != null) {
            options.put("zIndex", symbol.getSymbolSortKey());
        }
        if (symbol.getPosition() != null) {
            List<Double> geometry = new ArrayList<>();
            geometry.add(symbol.getPosition().getLatitude());
            geometry.add(symbol.getPosition().getLongitude());
            options.put("geometry", geometry);
        }
        options.put("draggable", symbol.isDraggable());
        data.put("options", options);
        return data;
    }

    static Map<String, Object> convertLineOptions(String id, Line line) {
        Map<String, Object> data = new HashMap<>();
        data.put("id", id);
        Map<String, Object> options = new HashMap<>();
        if (line.getLineJoin() != null) {
            options.put("lineJoin", line.getLineJoin());
        }
        if (line.getLineOpacity() != null) {
            options.put("lineOpacity", line.getLineOpacity());
        }
        if (line.getLineColor() != null) {
            options.put("lineColor", line.getLineColor());
        }
        if (line.getLineWidth() != null) {
            options.put("lineWidth", line.getLineWidth());
        }
        if (line.getLineGapWidth() != null) {
            options.put("lineGapWidth", line.getLineGapWidth());
        }
        if (line.getLineOffset() != null) {
            options.put("lineOffset", line.getLineOffset());
        }
        if (line.getLineBlur() != null) {
            options.put("lineBlur", line.getLineBlur());
        }
        if (line.getLinePattern() != null) {
            options.put("linePattern", line.getLinePattern());
        }
        List<List<Double>> geometry = new ArrayList<>();
        for (com.mapmyindia.sdk.geojson.Point lineString : line.getGeometry().coordinates()) {
            List<Double> points = new ArrayList<>();
            points.add(lineString.latitude());
            points.add(lineString.longitude());
            geometry.add(points);
        }
        options.put("geometry", geometry);
        options.put("draggable", line.isDraggable());
        data.put("options", options);
        return data;
    }

    static Map<String, Object> convertCircleOptions(String id, Circle circle) {
        Map<String, Object> data = new HashMap<>();
        data.put("id", id);
        Map<String, Object> options = new HashMap<>();
        if (circle.getCircleRadius() != null) {
            options.put("circleRadius", circle.getCircleRadius());
        }
        if (circle.getCircleColor() != null) {
            options.put("circleColor", circle.getCircleColor());
        }
        if (circle.getCircleBlur() != null) {
            options.put("circleBlur", circle.getCircleBlur());
        }
        if (circle.getCircleOpacity() != null) {
            options.put("circleOpacity", circle.getCircleOpacity());
        }
        if (circle.getCircleStrokeWidth() != null) {
            options.put("circleStrokeWidth", circle.getCircleStrokeWidth());
        }
        if (circle.getCircleStrokeColor() != null) {
            options.put("circleStrokeColor", circle.getCircleStrokeColor());
        }
        if (circle.getCircleStrokeOpacity() != null) {
            options.put("circleStrokeOpacity", circle.getCircleStrokeOpacity());
        }
        List<Double> geometry = new ArrayList<>();
        geometry.add(circle.getPosition().getLatitude());
        geometry.add(circle.getPosition().getLongitude());
        options.put("geometry", geometry);

        options.put("draggable", circle.isDraggable());
        data.put("options", options);
        return data;
    }

    static Map<String, Object> convertFillOptions(String id, Fill fill) {
        Map<String, Object> data = new HashMap<>();
        data.put("id", id);
        Map<String, Object> options = new HashMap<>();
        if (fill.getFillOpacity() != null) {
            options.put("fillOpacity", fill.getFillOpacity());
        }
        if (fill.getFillColor() != null) {
            options.put("fillColor", fill.getFillColor());
        }
        if (fill.getFillOutlineColor() != null) {
            options.put("fillOutlineColor", fill.getFillOutlineColor());
        }
        if (fill.getFillPattern() != null) {
            options.put("fillPattern", fill.getFillPattern());
        }
        List<List<List<Double>>> geometry = new ArrayList<>();
        for (List<LatLng> pointsList : fill.getPoints()) {
            List<List<Double>> geometryList = new ArrayList<>();
            for (LatLng latLng : pointsList) {
                List<Double> points = new ArrayList<>();
                points.add(latLng.getLatitude());
                points.add(latLng.getLatitude());
                geometryList.add(points);
            }
            geometry.add(geometryList);
        }
        options.put("geometry", geometry);
        options.put("draggable", fill.isDraggable());
        data.put("options", options);
        return data;
    }
    
    static MapmyIndiaGeoAnalyticsType getGeoAnalyticsType(int position) {
        MapmyIndiaGeoAnalyticsType[] types = new MapmyIndiaGeoAnalyticsType[]{
                MapmyIndiaGeoAnalyticsType.STATE,
                MapmyIndiaGeoAnalyticsType.DISTRICT,
                MapmyIndiaGeoAnalyticsType.SUB_DISTRICT,
                MapmyIndiaGeoAnalyticsType.TOWN,
                MapmyIndiaGeoAnalyticsType.CITY,
                MapmyIndiaGeoAnalyticsType.PINCODE,
                MapmyIndiaGeoAnalyticsType.WARD,
                MapmyIndiaGeoAnalyticsType.LOCALITY,
                MapmyIndiaGeoAnalyticsType.PANCHAYAT,
                MapmyIndiaGeoAnalyticsType.BLOCK,
                MapmyIndiaGeoAnalyticsType.VILLAGE,
                MapmyIndiaGeoAnalyticsType.SUB_LOCALITY,
                MapmyIndiaGeoAnalyticsType.SUB_SUB_LOCALITY
        };
        return types[position];
    }

    static List<MapmyIndiaGeoAnalyticsRequest> getGeoAnalyticsRequest(Object o) {
        Map<?, ?> map = toMap(o);
        List<MapmyIndiaGeoAnalyticsRequest> list = new ArrayList<>();
        String geoBoundType = toString(map.get("geoboundType"));
        String attribute = toString(map.get("attribute"));
        List<?> propertyNames = toList(map.get("propertyNames"));
        String[] propertyNamesArray = new String[propertyNames.size()];
        for (int i = 0; i < propertyNames.size(); i++) {
            propertyNamesArray[i] = toString(propertyNames.get(i));
        }
        Boolean transparent = map.containsKey("transparent") ? toBoolean(map.get("transparent")) : null;
        String query = map.containsKey("query") ? toString(map.get("query")) : null;
        List<?> requestList = toList(map.get("geoAnalyticsRequest"));
        for (Object requestObject : requestList) {
            Map<?, ?> request = toMap(requestObject);
            MapmyIndiaGeoAnalyticsRequest.Builder builder = MapmyIndiaGeoAnalyticsRequest.builder()
//                    .attribute(attribute)
                    .geoboundType(geoBoundType)
                    .geoBound(toString(request.get("geobound")))
                    .propertyNames(propertyNamesArray);
            if (transparent != null) {
                builder.transparent(transparent);
            }
            if (request.containsKey("style")) {
                builder.style(getGeoAnalyticsStyle(request.get("style")));
            }
            if (query != null) {
                builder.query(query);
            }
            if (attribute != null) {
                builder.attribute(attribute);
            }
            list.add(builder.build());
        }
        return list;
    }

    private static GeoAnalyticsAppearanceOption getGeoAnalyticsStyle(Object o) {
        Map<?,?> map = toMap(o);
        GeoAnalyticsAppearanceOption option = new GeoAnalyticsAppearanceOption();
        if(map.containsKey("fillColor")) {
            String fillColor = toString(map.get("fillColor"));
            if (fillColor != null && fillColor.contains("#")) {
                fillColor = fillColor.substring(1);
            }
            option.fillColor(fillColor);
        }
        if(map.containsKey("fillOpacity")) {
            option.fillOpacity(toDouble(map.get("fillOpacity")));
        }
        if(map.containsKey("labelColor")) {
            String labelColor = toString(map.get("labelColor"));
            if (labelColor != null && labelColor.contains("#")) {
                labelColor = labelColor.substring(1);
            }
            option.labelColor(labelColor);
        }
        if(map.containsKey("labelSize")) {
            option.labelSize(toInt(map.get("labelSize")));
        }
        if(map.containsKey("pointSize")) {
            option.pointSize(toInt(map.get("pointSize")));
        }
        if(map.containsKey("strokeColor")) {
            String strokeColor = toString(map.get("strokeColor"));
            if (strokeColor != null && strokeColor.contains("#")) {
                strokeColor = strokeColor.substring(1);
            }
            option.strokeColor(strokeColor);
        }
        if (map.containsKey("strokeWidth")) {
            option.strokeWidth(toDouble(map.get("strokeWidth")));
        }
        return option;
    }

    static List<Map<String, Object>> getGeoAnalyticsDetails(List<MapmyIndiaGeoAnalyticsDetail> details) {
        List<Map<String, Object>> mapList = new ArrayList<>();
        for (MapmyIndiaGeoAnalyticsDetail detail : details) {
            Map<String, Object> map = new HashMap<>();
            map.put("properties", detail.getProperties());
            map.put("type", getGeoAnalyticsType(detail.getType()));
            mapList.add(map);
        }
        return mapList;
    }

    private static int getGeoAnalyticsType(String type) {
        if(type == null) {
            return 0;
        }
        switch (type) {
            case "district":
                return 1;
            case "subdistrict":
                return 2;
            case "town":
                return 3;
            case "city":
                return 4;
            case "pincode":
                return 5;
            case "ward":
                return 6;
            case "locality":
                return 7;
            case "panchayat":
                return 8;
            case "block":
                return 9;
            case "village":
                return 10;
            case "sublocality":
                return 11;
            case "sub_sublocality":
                return 12;
            default:
                return 0;
        }
    }
}